// Avisynth v2.5.  Copyright 2002 Ben Rudiak-Gould et al.
// http://www.avisynth.org

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .
//
// Linking Avisynth statically or dynamically with other modules is making a
// combined work based on Avisynth.  Thus, the terms and conditions of the GNU
// General Public License cover the whole combination.
//
// As a special exception, the copyright holders of Avisynth give you
// permission to link Avisynth with independent modules that communicate with
// Avisynth solely through the interfaces defined in avisynth.h, regardless of the license
// terms of these independent modules, and to copy and distribute the
// resulting combined work under terms of your choice, provided that
// every copy of the combined work is accompanied by a complete copy of
// the source code of Avisynth (the version of Avisynth used to produce the
// combined work), being distributed under the terms of the GNU General
// Public License plus this exception.  An independent module is a module
// which is not derived from or based on Avisynth, such as 3rd-party filters,
// import and export plugins, or graphical user interfaces.

#include "stdafx.h"

#include "resample_functions.h"


/*******************************************
   ***************************************
   **  Helper classes for resample.cpp  **
   ***************************************
 *******************************************/

/***********************
 *** Spline100 filter ***
 ***********************/

double Spline100Filter::f(double value, double *coefficients) {
  value = fabs(value);

  if        (value < 1.0) {
    return (( 61.0/51.0 * (value    ) - 9893.0/4505.0) * (value    ) -    1.0/13515.0) * (value    ) + 1.0;
  } else if (value < 2.0) {
    return ((-10.0/17.0 * (value-1.0) +  1254.0/901.0) * (value-1.0) -    724.0/901.0) * (value-1.0);
  } else if (value < 3.0) {
    return ((  8.0/51.0 * (value-2.0) - 1672.0/4505.0) * (value-2.0) + 2896.0/13515.0) * (value-2.0);
  } else if (value < 4.0) {
    return (( -2.0/51.0 * (value-3.0) +  418.0/4505.0) * (value-3.0) -  724.0/13515.0) * (value-3.0);
  } else if (value < 5.0) {
    return (( 1.0/153.0 * (value-4.0) - 209.0/13515.0) * (value-4.0) +  362.0/40545.0) * (value-4.0);
  }
  return 0.0;
}


/***********************
 *** Spline144 filter ***
 ***********************/

double Spline144Filter::f(double value, double *coefficients) {
  value = fabs(value);

  if        (value < 1.0) {
    return (( 683.0/571.0 * (value    ) - 1240203.0/564719.0) * (value    ) -      3.0/564719.0) * (value    ) + 1.0;
  } else if (value < 2.0) {
    return ((-336.0/571.0 * (value-1.0) +  786240.0/564719.0) * (value-1.0) - 453936.0/564719.0) * (value-1.0);
  } else if (value < 3.0) {
    return ((  90.0/571.0 * (value-2.0) -  210600.0/564719.0) * (value-2.0) + 121590.0/564719.0) * (value-2.0);
  } else if (value < 4.0) {
    return ((- 24.0/571.0 * (value-3.0) +   56160.0/564719.0) * (value-3.0) -  32424.0/564719.0) * (value-3.0);
  } else if (value < 5.0) {
    return ((   6.0/571.0 * (value-4.0) -   14040.0/564719.0) * (value-4.0) +   8106.0/564719.0) * (value-4.0);
  } else if (value < 6.0) {
    return ((  -1.0/571.0 * (value-5.0) +    2340.0/564719.0) * (value-5.0) -   1351.0/564719.0) * (value-5.0);
  }
  return 0.0;
}


/***********************
 *** Spline filter ***
 ***********************/

double *Spline100Filter::coeff(int taps) {
  return 0;
}

double *Spline144Filter::coeff(int taps) {
  return 0;
}

double *SplineFilter::coeff(int taps) {
  double* coefficients = (double*)malloc((4*taps+1)*sizeof(double));
/*
  a(j) = a(2*taps-j) (this holds for b(j), c(j) and d(j) as well)
  c2 = [taps, a(taps+1), b(taps+1), c(taps+1), d(taps+1),
			  a(taps+2), b(taps+2), c(taps+2), d(taps+2),
			  ......................
			  a(2*taps), b(2*taps), c(2*taps), d(2*taps)]
*/
  
  // Lz = f, z = Ux, n = 2*taps+1
  // z = [z[0]; ...; z[n-1]]
  // f = [f[0]; ...; f[n-1]]
  // x = [x[1]; ...; x[n]] (that's not consistent, but i need that x[0]=x[n+1]=0)
  // LU = [4 1 0 0 ... ]
  //	  [1 4 1 0 ... ]
  //	  [0 1 4 1 ... ]
  //	   ...
  //	  [0 ...  1 4 1]
  //	  [0 ...  0 1 4]
  double* y = (double*)malloc((2*taps+1)*sizeof(double));
  double* f = (double*)malloc((2*taps-1)*sizeof(double));
  double* w = (double*)malloc((2*taps)*sizeof(double));
  double* z = (double*)malloc((2*taps)*sizeof(double));
  double* x = (double*)malloc((2*taps+1)*sizeof(double));

  for (int j=0; j<(2*taps+1); ++j) {
	y[j] = 0;
  }
  y[taps] = 1.0;

  for (j=0; j<(2*taps-1); ++j) {
	f[j] = 0;
  }

  if (taps>1) {
    f[taps-2] = 6.0;
    f[taps] = 6.0;
  }

  f[taps-1] = -12.0;

  w[0] = 4.0;
  z[0] = f[0]/w[0];
  for (j=1; j<(2*taps); ++j) {
	w[j] = 4.0 - 1.0/w[j-1];
	z[j] = (f[j] - z[j-1])/w[j];
  }

  x[0] = 0;
  x[2*taps] = 0;

//  x[2*taps-1] = z[2*taps-2];
//  for (j=2*taps-2; j>0; --j) {
  for (j=2*taps-1; j>0; --j) {
	x[j] = z[j-1] - x[j+1]/w[j-1];
  }

/*
  j=1:2*taps
  a(j) = (x(j) - x(j-1))/6
  b(j) = x(j-1)/2
  c(j) = [y(j) - y(j-1)] - [x(j) + 2*x(j-1)]/6
  d(j) = y(j-1)
  the values (a,b,c,d) for j is the same as for 2*taps-j
*/

  coefficients[0] = taps;

  for (j=taps; j<(2*taps); ++j) {
	coefficients[4*(j-taps)+1] = (x[j+1] - x[j])/6.0; // a(j)
	coefficients[4*(j-taps)+2] = x[j]/2.0; // b(j)
	coefficients[4*(j-taps)+3] = (y[j+1] - y[j]) - (x[j+1] + 2*x[j])/6.0; // c(j)
	coefficients[4*(j-taps)+4] = y[j]; // d(j)
  }

  free(y);
  free(f);
  free(w);
  free(z);
  free(x);

  return coefficients;
}

SplineFilter::SplineFilter(int t = 4) {
   taps = (double)(max(1,min(100,t)));
}


double SplineFilter::f(double value, double *coefficients) {
  value = fabs(value);

  for (int j=1; j<coefficients[0]+1; ++j) {
    if (value < j) {
      return ((coefficients[4*(j-1)+1] * (value-j+1) + coefficients[4*(j-1)+2]) * (value-j+1) + coefficients[4*(j-1)+3]) * (value-j+1) + coefficients[4*(j-1)+4];
	}
  }

  return 0.0;
}

/*
double SplineFilter::f(double value, double *coefficients) { // Spline64
  value = fabs(value);

  if        (value < 1.0) {
    return (116.0/97.0 * value - 213.0/97.0) * value * value + 1.0;
  } else if (value < 2.0) {
    return ((-57.0/97.0 * value + 306.0/97.0) * value - 519.0/97.0) * value + 270.0/97.0;
  } else if (value < 3.0) {
    return ((15.0/97.0 * value - 126.0/97.0) * value + 345.0/97.0) * value - 306.0/97.0;
  } else if (value < 4.0) {
    return ((-3.0/97.0 * value + 36.0/97.0) * value - 141.0/97.0) * value + 180.0/97.0;
  }

  return 0.0;
}
*/



/******************************
 **** Resampling Patterns  ****
 *****************************/

int* GetResamplingPatternRGB( int original_width, double subrange_start, double subrange_width,
                              int target_width, ResamplingFunction* func, IScriptEnvironment* env )
/**
  * This function returns a resampling "program" which is interpreted by the 
  * FilteredResize filters.  It handles edge conditions so FilteredResize    
  * doesn't have to.  
 **/
{
  double scale = double(target_width) / subrange_width;
  double filter_step = min(scale, 1.0);
  double filter_support = func->support() / filter_step;
  int fir_filter_size = int(ceil(filter_support*2));

  double *x = func->coeff(func->support());

  int* result = (int*) _aligned_malloc((1 + target_width*(1+fir_filter_size)) * 4, 64);

  int* cur = result;
  *cur++ = fir_filter_size;

  double pos_step = subrange_width / target_width;
  // the following translates such that the image center remains fixed
  double pos;

  if (original_width <= filter_support) {
    env->ThrowError("Resize: Source image too small for this resize method");
  }

  if (fir_filter_size == 1) // PointResize
    pos = subrange_start;
  else
    pos = subrange_start + ((subrange_width - target_width) / (target_width*2));

  for (int i=0; i<target_width; ++i) {
    int end_pos = int(pos + filter_support);

    if (end_pos > original_width-1)
      end_pos = original_width-1;

    int start_pos = end_pos - fir_filter_size + 1;

    if (start_pos < 0)
      start_pos = 0;

    *cur++ = start_pos;

    // the following code ensures that the coefficients add to exactly FPScale
    double total = 0.0;

    // Ensure that we have a valid position
    double ok_pos = max(0.0,min(original_width-1,pos));

    for (int j=0; j<fir_filter_size; ++j) {  // Accumulate all coefficients
      total += func->f((start_pos+j - ok_pos) * filter_step, x);
    }

    if (total == 0.0) {
      // Shouldn't happend for valid positions.
#ifdef _DEBUG
      env->ThrowError("Resizer: [Internal Error] Got Zero Coefficient");
#endif
      total = 1.0;
    }

    double total2 = 0.0;

    for (int k=0; k<fir_filter_size; ++k) {
      double total3 = total2 + func->f((start_pos+k - ok_pos) * filter_step, x) / total;
      *cur++ = int(total3*FPScale+0.5) - int(total2*FPScale+0.5);
      total2 = total3;
    }

    pos += pos_step;
  }

  return result;
}


int* GetResamplingPatternYUV( int original_width, double subrange_start, double subrange_width,
                              int target_width, ResamplingFunction* func, bool luma, BYTE *temp,
                              IScriptEnvironment* env )
/**
  * Same as with the RGB case, but with special
  * allowances for YUV-MMX code
 **/
{
  double scale = double(target_width) / subrange_width;
  double filter_step = min(scale, 1.0);
  double filter_support = func->support() / filter_step;
  int fir_filter_size = int(ceil(filter_support*2));
  int fir_fs_mmx = (fir_filter_size / 2) +1;  // number of loops in MMX code
  int target_width_a=(target_width+15)&(~15);

  double *x = func->coeff(func->support());  
  
  int* result = luma ?
                (int*) _aligned_malloc(2*4 + target_width_a*(1+fir_fs_mmx)*8, 64) :
                (int*) _aligned_malloc(2*4 + target_width_a*(1+fir_filter_size)*8, 64);

  int* cur[2] = { result +2, result +3 };
  *result = luma ? fir_fs_mmx : fir_filter_size;

  double pos_step = subrange_width / target_width;
  // the following translates such that the image center remains fixed
  double pos;

  if (fir_filter_size == 1) // PointResize
    pos = subrange_start;
  else
    pos = subrange_start + ((subrange_width - target_width) / (target_width*2));

  if (original_width <= filter_support) {
    env->ThrowError("Resize: Source image too small for this resize method");
  }

  for (int i=0; i<target_width_a; ++i) {
    int end_pos = int(pos + filter_support);

    if (end_pos > original_width-1)  //This will ensure that the filter will not end beyond the end of the line.
      end_pos = original_width-1;

    int start_pos = end_pos - fir_filter_size + 1;  // Calculate where to start, so we don't end outside the line.

    if (start_pos < 0)  // Did we get too far back?
      start_pos = 0;

    int ii = luma ? i&1 : 0;

    *(cur[ii]) = luma ?   // Write offset of first pixel.
                 (int)(temp + (start_pos & -2) * 2) :
                 (int)(temp + start_pos * 8);

    cur[ii] += 2;

    // the following code ensures that the coefficients add to exactly FPScale
    double total = 0.0;

    // Ensure that we have a valid position
    double ok_pos = max(0.0,min(original_width-1, pos)); 

    for (int j=0; j<fir_filter_size; ++j) {  // Accumulate all coefficients
      total += func->f((start_pos + j - ok_pos) * filter_step, x);
    }

    if (total == 0.0) {
      // Shouldn't happend for valid positions.
#ifdef _DEBUG
      env->ThrowError("Resizer: [Internal Error] Got Zero Coefficient");
#endif
      total = 1.0;
    }

    double total2 = 0.0;
    int oldCoeff = 0;

    for (int k=0; k<fir_filter_size; ++k) {
      double total3 = total2 + func->f((start_pos+k - ok_pos) * filter_step, x) / total;
      int coeff = int(total3*FPScale+0.5) - int(total2*FPScale+0.5);
      total2 = total3;

      if (luma) {
        if ((k + start_pos) & 1) {
          *(cur[ii]) = (coeff << 16) + (oldCoeff & 0xFFFF);
          cur[ii] += 2;
        } else
          oldCoeff = coeff;
      } else {
        *(cur[0]) = coeff;
        cur[0] += 1;
        *(cur[0]) = coeff;
        cur[0] += 1;
      }
    }

    if (luma) {
      if ((start_pos + fir_filter_size) & 1) {
        *(cur[ii]) = 0 + (oldCoeff & 0xFFFF);
        cur[ii] += 2;
      } else
        if ((fir_filter_size & 1) == 0) {
          *(cur[ii]) = 0;
          cur[ii] += 2;
        }
    }

    pos += pos_step;
  }

  return result;
}
