// Avisynth v2.5.  Copyright 2002 Ben Rudiak-Gould et al.
// http://www.avisynth.org

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .
//
// Linking Avisynth statically or dynamically with other modules is making a
// combined work based on Avisynth.  Thus, the terms and conditions of the GNU
// General Public License cover the whole combination.
//
// As a special exception, the copyright holders of Avisynth give you
// permission to link Avisynth with independent modules that communicate with
// Avisynth solely through the interfaces defined in avisynth.h, regardless of the license
// terms of these independent modules, and to copy and distribute the
// resulting combined work under terms of your choice, provided that
// every copy of the combined work is accompanied by a complete copy of
// the source code of Avisynth (the version of Avisynth used to produce the
// combined work), being distributed under the terms of the GNU General
// Public License plus this exception.  An independent module is a module
// which is not derived from or based on Avisynth, such as 3rd-party filters,
// import and export plugins, or graphical user interfaces.

#ifndef __Resample_Functions_H__
#define __Resample_Functions_H__

#include "internal.h"

// Original value: 65536
// 2 bits sacrificed because of 16 bit signed MMX multiplication
const int FPScale = 16384; // fixed point scaler

// 09-14-2002 - Vlad59 - Lanczos3Resize - Constant added
#define M_PI 3.14159265358979323846



/*******************************************
   ***************************************
   **  Helper classes for resample.cpp  **
   ***************************************
 *******************************************/


class ResamplingFunction 
/**
  * Pure virtual base class for resampling functions
  */
{
public:
  virtual double *coeff(int taps) = 0;
  virtual double f(double x, double *coefficients) = 0;
  virtual double support() = 0;
};

class Spline100Filter : public ResamplingFunction
/**
  * Spline100 is like Spline64, except that it uses 10x10=100 pixels.
 **/
{
public:
	double *coeff(int taps);
	double f(double x, double *coefficients);
	double support() { return 5.0; };

private:
};

class Spline144Filter : public ResamplingFunction
/**
  * Spline144 is like Spline64, except that it uses 12x12=144 pixels.
 **/
{
public:
	double *coeff(int taps);
	double f(double x, double *coefficients);
	double support() { return 6.0; };

private:
};

class SplineFilter : public ResamplingFunction
/**
  * Spline is like Spline64, except using sensible coefficients
 **/
{
public:
	double *coeff(int taps);
	SplineFilter(int taps);
	double f(double x, double *coefficients);
	double support() { return taps; };

private:
  double taps;
};


int* GetResamplingPatternRGB(int original_width, double subrange_start, double subrange_width,
                                    int target_width, ResamplingFunction* func, IScriptEnvironment* env);

int* GetResamplingPatternYUV(int original_width, double subrange_start, double subrange_width,
                                    int target_width, ResamplingFunction* func, bool luma, BYTE *temp,
                                    IScriptEnvironment* env);


#endif  // __Reample_Functions_H__
